#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef __SH4__
#include <pthread.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#endif

// RC server
#include <arpa/inet.h>        /*  inet (3) funtions         */
#include <errno.h>

/*  Global constants  */
#define SERVER_PORT          (2345)

ssize_t Readline(int sockd, void *vptr, size_t maxlen);

void *remote_thread( void *v )
{
	FILE *hlog;

#ifdef __SH4__
 	hlog = fopen( "/proc/self/fd/1", "a+" );
#else 
 	hlog = fopen( "/mtd_ram/remote.log", "a+" );
#endif

	fprintf( hlog, "remote_thread started: pInstance - %p\n", v );
	fflush(hlog);
	
	int       list_s;
	int       conn_s;
	struct    sockaddr_in servaddr;
	char      buffer[1024]; 

#ifdef __SH4__
  int *piRcKeyQueueId= dlsym(RTLD_DEFAULT, "_ZN9KeyCommon15m_iRcKeyQueueIdE");
  int iRcKeyQueueId= 0;

  if (piRcKeyQueueId) iRcKeyQueueId= *piRcKeyQueueId;
	fprintf(hlog, "RcKeyQueueId %d\n", iRcKeyQueueId);
#endif

	if ( (list_s = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
	{
		fprintf(hlog, "Error creating listening socket.\n");
		fclose(hlog);
		return NULL;
    	}

	int on = 1;
	setsockopt(list_s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(int));

	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family      = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port        = htons(SERVER_PORT);

	if ( bind(list_s, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0 ) 
	{
		fprintf(hlog, "Error calling bind()\n");
		fclose(hlog);
		return NULL;
	}

	if ( listen(list_s, 10) < 0 ) 
	{
		fprintf(hlog, "Error calling listen()\n");
		fclose(hlog);
		return NULL;
	}

	while (1)
	{
		if ( (conn_s = accept(list_s, NULL, NULL) ) < 0 ) {
		    fprintf( hlog, "ECHOSERV: Error calling accept()\n");
		    break;
		}

		memset( buffer, 0, 1024 );
		Readline(conn_s, buffer, 1023);

		if ( buffer[0] == 'Q' )
		{
		    fprintf( hlog, "exit..\n");
			close(conn_s);
			break;
		}

		int key = strtol( buffer, NULL, 0 );

		fprintf( hlog, "received key: %d\n", key );
		fflush(hlog);

#ifdef __SH4__
		struct msgbuf {
      long mtype;
      long data[3];
    } msg;
    msg.mtype= 1;
    msg.data[0]= -3; // -3..simulated rc, -6 real rc
    msg.data[1]= 0;
    msg.data[2]= key;
    msgsnd(iRcKeyQueueId, &msg, 12, 0);
#else 
		asm volatile(
			"MOV	R0,%0\n"
			"MOV 	R1,%1\n"
			"BL	_ZN14SsKeyInputBase7SendKeyEi\n"
			:
			: "r" (v), "r" (key) );
#endif

		if ( close(conn_s) < 0 ) 
		{
		    fprintf(hlog, "Error calling close()\n");
		    break;
		}
	}

	fclose(hlog);
	close(list_s);
	return NULL;
}

ssize_t Readline(int sockd, void *vptr, size_t maxlen) {
    ssize_t n, rc;
    char    c, *buffer;

    buffer = vptr;

    for ( n = 1; n < maxlen; n++ ) {
	
	if ( (rc = read(sockd, &c, 1)) == 1 ) {
	    *buffer++ = c;
	    if ( c == '\n' )
		break;
	}
	else if ( rc == 0 ) {
	    if ( n == 1 )
		return 0;
	    else
		break;
	}
	else {
	    if ( errno == EINTR )
		continue;
	    return -1;
	}
    }

    *buffer = 0;
    return n;
}

#ifdef __SH4__
///////////////////////////////////////////////////////////////////////////////
volatile int thr_cancel= 0;
pthread_t thr;
void *status;

void _init(void)
{
  pthread_create(&thr, NULL, remote_thread, NULL);
}

int Game_Main(const char *path, const char *udn)
{
  // just increment usage count to prevent unload at Game_Main exit
  char filename[256];
  sprintf(filename, "%s%s", path, "libRemote.so");
  dlopen(filename, RTLD_LAZY);
  return 0;
}

///////////////////////////////////////////////////////////////////////////////

void _fini(void)
{
  pthread_cancel(thr);
  pthread_join(thr, &status);
}
#endif
